# -*- coding: utf-8 -*-
#    Copyright (C) 2007 Stewart Adam
#    This file is part of fwbackups.

#    fwbackups is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.

#    fwbackups is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with fwbackups; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
"""
The part of fwbackup that manipulates the actual files.
"""
import os
import os.path
import time
import logging

from fwbackups.i18n import _
from fwbackups.const import *
from fwbackups.config import BackupSetConf, OneTimeConf
from fwbackups import utils, shutil_modded, fwlogger

class EngineError(Exception):
    """ Engine's error """
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)

class Common:
    """ Common defs for all restore / backup / onetime """
    def __init__(self, logger=None):
        # Initialize a logger if none passed
        if not logger:
            self.logger, self.handler = fwlogger._setupLogger(\
                                                              level=logging.INFO, \
                                                              printToo=True\
                                                              )
        else:
            self.logger = logger
        # We're not cancelling...
        self.tocancel = False
        # Totals: None set.
        self.current, self.total  = [0,0]
        # Execute commands with this environment.
        self.environment = {'PATH':os.getenv('PATH')}
        
    def getStatus(self):
        return self.current, self.total

    def cancel(self):
        self.tocancel = True
        self.logger.logmsg('INFO', _('Cancel request sent, waiting for current \
command to finish'))

    def onError(self, func, path, exc_info):
        self.logger.logmsg('WARNING', _('Error removing old file \'%s\'.') % path)
        self.logger.logmsg('WARNING', _('Output:\n%s') % ('func:%s\nexc_info:%s' \
% (func, exc_info)))
        pass

    def ifCancel(self):
        if self.tocancel == True:
            self.logger.logmsg('INFO', _('Cancel request recieved, cleaning up'))
            sys.exit(0)

class Restore(Common):
    def __init__(self, source, dest, logger=None):
        Common.__init__(self, logger)
        self.source = source
        self.dest = dest
        self.restoredest = utils.ConvertPath('%s/%s-%s' % (self.dest,\
            _('Restore'), time.strftime('%Y-%m-%d')))
        self.logger.logmsg('INFO', _('Starting restore operation'))

    def ifCancel(self):
        if self.tocancel == True:
            self.logger.logmsg('INFO', _('Cancel request recieved, cleaning up'))
            try:
                shutil_modded.rmtree(self.restoredest)
            except:
                pass
            time.sleep(1)
            sys.exit(0)

    def restore(self):
        """ Restores a backup """
        # ungzip if needed
        toremove = 0
        if self.source.endswith('tar.gz'):
            toremove = 1
            self.ifCancel()
            self.logger.logmsg('INFO', _('Decompressing the backup archive'))
            retval, stdout, stderr = utils.execute('gzip -dc \'%s\' > %s%sfwbackups-extracted.tar' % \
                                                   (\
                                                    self.source, \
                                                    pathsep.join(self.source.split(pathsep)[:-1]), \
                                                    pathsep\
                                                    ), \
                                                   env=self.environment, shell=True)
            if retval != os.EX_OK:
                self.logger.logmsg('WARNING', _('Error decompressing the backup archive!'))
                raise EngineError(_('Error decompressing the backup archive!'))
            temp = self.source.split(pathsep)[:-1]
            temp.append('fwbackups-extracted.tar')
            self.source = pathsep.join(temp)
        try:
            if not os.path.exists(self.dest):
                os.mkdir(self.dest)
        except:
            self.logger.logmsg('ERROR', _('The backup destination \'%s\' is non\
-existant and could not be created.') % self.dest)
            raise EngineError(_('The backup destination \'%s\' is non-existant \
and could not be created.') % self.dest)
        if not utils.CheckPerms(self.dest):
            self.logger.logmsg('ERROR', _('You do not have read and write \
permissions on the backup destination \'%s\'') % self.dest)
            raise EngineError(_('You do not have read and write permissions \
on the backup destination \'%s\'') % self.dest)

        self.ifCancel()
        if os.path.exists(self.restoredest):
            self.logger.logmsg('INFO', _(' \'%s\' already exists! Removing \
first...') % (self.restoredest))
            shutil_modded.rmtree(path=self.restoredest, onerror=self.onError)
        os.mkdir(self.restoredest)
        
        # tar gz, tar
        if self.source.endswith('.tar'):
            os.chdir(self.dest)
            command = 'cd \'%s\';tar xf \'%s\' -p --same-owner' % (self.restoredest, self.source)
        # rsync
        else:
            command = 'rsync -g -o -p -r \'%s\'/* \'%s\'' % (self.source, self.restoredest)
        self.ifCancel()
        retval, stdout, stderr = utils.execute(command, env=self.environment, shell=True)
        if retval != os.EX_OK:# and retval != 2:
            self.logger.logmsg('ERROR','Error while restoring certain files or folders. The restore may or may not be incomplete.')
            self.logger.logmsg('ERROR', _('Output:\n%s') % ('Return value: %s\nstdout: %s\nstderr: %s' %\
                (\
                 retval, \
                 ''.join(stdout.readlines()), \
                 ''.join(stderr.readlines())
                 )))
        if toremove == 1:
            os.remove(self.source)

        # All done!
        self.logger.logmsg('INFO', _('Finished restore operation'))


class OneTime(Common):
    """ One-time backup """
    def __init__(self, confloc, logger=None):
        Common.__init__(self, logger)
        self.confloc = confloc
        self.logger.logmsg('INFO', _('Starting one-time backup'))
        self.set = OneTimeConf()
        self.folderdest = self.set.get('Options', 'Destination')
        self.backupdest = utils.ConvertPath('%s/%s-OneTime-%s' % (self.folderdest,\
            _('Backup'), time.strftime('%Y-%m-%d')))
        self.engine = self.set.get('Options', 'Engine')
        self.recursive = self.set.get('Options', 'Recursive')
        self.pkglist = self.set.get('Options', 'PkgListsToFile')
        self.diskinfo = self.set.get('Options', 'DiskInfoToFile')
        self.dohidden = self.set.get('Options', 'BackupHidden')

    def ifCancel(self):
        if self.tocancel == True:
            self.logger.logmsg('INFO', _('Cancel request recieved, cleaning up'))
            try:
                shutil_modded.rmtree(self.backupdest)
                shutil_modded.rmtree(self.backupdest + '.tar')
                shutil_modded.rmtree(self.backupdest + '.tar.gz')
            except:
                pass
            time.sleep(1)
            sys.exit(0)

    def _prepareDestination(self):
        """ Prepare the destination folder """
        try:
            if not os.path.exists(self.folderdest):
                os.mkdir(self.folderdest)
            self.logger.logmsg('DEBUG', _('The backup destination \'%s\' was \
non-existant and and has been successfully created.') % self.folderdest)
        except:
            self.logger.logmsg('ERROR', _('The backup destination \'%s\' is non\
-existant and could not be created.') % self.folderdest)
            raise EngineError(_('The backup destination \'%s\' is non-existant\
 and could not be created.') % self.folderdest)
        if not utils.CheckPerms(self.folderdest):
            self.logger.logmsg('ERROR', _('You do not have read and write\
 permissions on the backup destination \'%s\'') % self.folderdest)
            raise EngineError(_('You do not have read and write permissions\
 on the backup destination \'%s\'') % self.folderdest)

    def _prepareFullDestination(self):
        """ Prepare the backup destination """
        # Full destination exists?
        self.ifCancel()
        if os.path.exists(self.backupdest):
            self.logger.logmsg('INFO', _('\'%s\' already exists! It is being\
 removed so that the backup can continue.') % (self.backupdest))
            shutil_modded.rmtree(path=utils.ConvertPath(self.backupdest), \
onerror=self.onError)
        if os.path.exists('%s.gz' % self.backupdest):
            self.logger.logmsg('INFO', _('\'%s\' already exists! It is being\
 removed so that the backup can continue.') % ('%s.gz' % self.backupdest))
            shutil_modded.rmtree(path=utils.ConvertPath('%s.gz' % \
self.backupdest), onerror=self.onError)

    def backup(self):
        """ One Time Backup """
        # Package lists?
        if self.pkglist == '1':
            self.ifCancel()
            utils.execute('rpm -qa --qf "%%{NAME}\n" > "/tmp/%s.txt"' % \
                          _('Package list, names only'),\
                env=self.environment, shell=True)
            utils.execute('rpm -qa > "/tmp/%s.txt"' % _('Package list'), \
env=self.environment, shell=True)
        # Disk info?
        if self.diskinfo == '1':
            self.ifCancel()
            utils.execute('fdisk -l > "/tmp/%s.txt"' % _('Disk Information'), \
env=self.environment, shell=True)

        # IF and only IF tar / tar.gz, add .tar to since
        # it's a file, not a folder the full destination
        if self.engine.startswith('tar'):
            self.backupdest += '.tar'

        self._prepareDestination()
        self._prepareFullDestination()

        # Commands, etc based on engine
        if self.engine.startswith('tar'):
            command = 'tar rf \'%s\'' % (self.backupdest)
            os.chdir('/tmp')
            if self.pkglist == '1':
                utils.execute('%s "/tmp/%s.txt" "/tmp/%s.txt"' % (command, \
_('Package list'), _('Package list, names only')), env=self.environment, \
shell=True, timeout=-1)
                utils.execute('rm "/tmp/%s.txt" "/tmp/%s.txt" -f' % \
(_('Package list'), _('Package list, names only')), env=self.environment, \
shell=True, timeout=-1)
            if self.diskinfo == '1':
                utils.execute('%s "/tmp/%s.txt"' % (command, _('Disk Information')),\
                    env=self.environment, shell=True, timeout=-1)
                utils.execute('rm "/tmp/%s.txt" -f' % _('Disk Information'),\
                    env=self.environment, shell=True, timeout=-1)
            if not self.recursive == '1':
                command += ' --no-recursion'
            if not self.dohidden == '1':
                command += ' --exclude=.*'

        elif self.engine == 'rsync':
            os.mkdir(self.backupdest)
            command = 'rsync -g -o -p'
            if self.recursive == '1':
                command += ' -r'
            if not self.dohidden == '1':
                command += ' --exclude=.*'
            if self.pkglist == '1':
                self.ifCancel()
                utils.execute('%s "/tmp/%s.txt" "/tmp/%s.txt" "%s"' % (command, \
_('Package list'), _('Package list, names only'), self.backupdest), \
env=self.environment, shell=True, timeout=-1)
                utils.execute('rm "/tmp/%s.txt" "/tmp/%s.txt" -f' % \
(_('Package list'), _('Package list, names only')), env=self.environment, \
shell=True, timeout=-1)
            if self.diskinfo == '1':
                self.ifCancel()
                utils.execute('%s "/tmp/%s.txt" "%s"' % (command, \
_('Disk Information'), self.backupdest), env=self.environment, shell=True, timeout=-1)
                utils.execute('rm "/tmp/%s.txt" -f' % _('Disk Information'), \
                              env=self.environment, shell=True, timeout=-1)

        # Get the list of paths...
        paths = []
        for i in self.set.options('Paths'):
            paths.append('\'%s\'' % self.set.get('Paths', i))
            self.total = self.total + 1

        # Now that the paths & commands are set up...
        for i in paths:
            self.ifCancel()
            self.current = self.current + 1
            self.logger.logmsg('DEBUG', _('Backing up path: %s') % i.split('\'')[1])
            if self.engine == 'rsync':
                i = '%s "%s"' % (i, self.backupdest)
            retval, stdout, stderr = utils.execute('%s %s' % (command, i), \
env=self.environment, shell=True)
            if retval != os.EX_OK and retval != 2:
                self.logger.logmsg('ERROR', _('Error backing up while path: %s') % i)
                self.logger.error(_('Output:n%s') % ('Return value: %s\nstdout: %s\nstderr: %s' % \
(retval, stdout.readlines(), stderr.readlines())))

        # gzip at end if needed
        if self.engine == 'tar.gz':
            self.ifCancel()
            self.logger.logmsg('DEBUG', _('Compresssing the backup archive'))
            retval, stdout, stderr = utils.execute('gzip \'%s\'' % \
self.backupdest, env=self.environment, shell=True)
            if retval != os.EX_OK:
                self.logger.logmsg('WARNING', _('An error occured while \
compressing the backup. Your backup is complete, but may not be (completely) \
compressed.'))
                self.logger.logmsg('ERROR', _('Output:\n%s') % ('Return value: %s\nstdout: %s\nstderr: %s' % \
(retval, stdout.readlines(), stderr.readlines())))

        # Backups are private!
        if self.engine != 'tar.gz':
            os.chmod(self.backupdest, 0711)
        else:
            os.chmod('%s.gz' % self.backupdest, 0711)

        # All done!
        self.logger.logmsg('INFO', _('Finished one-time backup'))



class Automated(Common):
    """ Automated backup of a set """
    def __init__(self, setname, logger=None):
        Common.__init__(self, logger)
        self.setname = setname
        self.logger.logmsg('INFO', _('Starting automatic backup of set \'%s\'') \
% self.setname)
        # We're using a set...
        self.set = BackupSetConf(self.setname)
        # Variables, extract set info.
        self.folderdest = self.set.get('Options', 'Destination')
        self.backupdest = utils.ConvertPath('%s/%s-%s-%s' % (self.folderdest, \
_('Backup'), self.setname, time.strftime('%Y-%m-%d')))
        self.engine = self.set.get('Options', 'Engine')
        self.command_before = self.set.get('Options', 'CommandBefore')
        self.command_after = self.set.get('Options', 'CommandAfter')
        if self.set.get('Options', 'Enabled') == 0:
            return
        self.tokeep = int(float(self.set.get('Options', 'OldToKeep')))
        self.recursive = self.set.get('Options', 'Recursive')
        self.pkglist = self.set.get('Options', 'PkgListsToFile')
        self.diskinfo = self.set.get('Options', 'DiskInfoToFile')
        self.dohidden = self.set.get('Options', 'BackupHidden')

    def ifCancel(self):
        if self.tocancel == True:
            self.logger.logmsg('INFO', _('Cancel request recieved, cleaning up'))
            try:
                shutil_modded.rmtree(self.backupdest)
                shutil_modded.rmtree(self.backupdest + '.tar')
                shutil_modded.rmtree(self.backupdest + '.tar.gz')
            except:
                pass
            time.sleep(1)
            sys.exit(0)

    def _prepareDestination(self):
        """ Prepare the destination folder """
        try:
            if not os.path.exists(self.folderdest):
                os.mkdir(self.folderdest)
        except:
            self.logger.logmsg('ERROR', _('The backup destination \'%s\' is non\
-existant and could not be created.') % self.folderdest)
            raise EngineError(_('The backup destination \'%s\' is non-existant \
and could not be created.') % self.folderdest)
        if not utils.CheckPerms(self.folderdest):
            self.logger.logmsg('ERROR', _('You do not have read and write \
permissions on the backup destination \'%s\'') % self.folderdest)
            raise EngineError(_('You do not have read and write \
permissions on the backup destination \'%s\'') % self.folderdest)

    def _prepareFullDestination(self):
        """ Prepare the backup destination """
        # Full destination exists?
        self.ifCancel()
        if os.path.exists(self.backupdest):
            self.logger.logmsg('INFO', _(' \'%s\' already exists! Removing \
first...') % (self.backupdest))
            shutil_modded.rmtree(path=self.backupdest, onerror=self.onError)
        if os.path.exists('%s.gz' % self.backupdest):
            self.logger.logmsg('INFO', _(' \'%s\' already exists! Removing \
first...') % ('%s.gz' % self.backupdest))
            shutil_modded.rmtree(path='%s.gz' % self.backupdest, onerror=self.onError)

    def backup(self):
        """ Backup a set """
        # Before command...
        if self.command_before != None and self.command_before != "":
            self.ifCancel()
            self.logger.logmsg('INFO', _('Executing \'Before\' command'))
            retval, stdout, stderr = utils.execute(self.command_before,\
 env=self.environment, shell=True)
            if retval != os.EX_OK:
                self.logger.logmsg('WARNING', _('Command failed; Please check \
your syntax to make sure it is correct.'))
        # Init main backup dir
        self._prepareDestination()
        # Init our backup, remove if existing
        self._prepareFullDestination()
        # Get list of old backups to remove...
        listing = os.listdir(self.folderdest)
        listing.sort()
        oldbackups = []
        for i in listing:
            if i.startswith('%s-%s-' % (_('Backup'), self.setname)):
                oldbackups.append(i)
        # ...And remove them.
        oldbackups.reverse()
        self.logger.logmsg('DEBUG', str(oldbackups))
        self.logger.logmsg('DEBUG', str(oldbackups[self.tokeep:]))
        for i in oldbackups[self.tokeep:]:
            self.ifCancel()
            self.logger.logmsg('DEBUG', _('Removing old backup \'%s\'') % i)
            shutil_modded.rmtree(path=utils.ConvertPath(\
                                                   '%s/%s' % (self.folderdest, i)), \
                                                   onerror=self.onError\
                                                   )
        
        # Package lists?
        if self.pkglist == '1':
            self.ifCancel()
            utils.execute('rpm -qa --qf "%%{NAME}\n" > "/tmp/%s.txt"' % \
_('Package list, names only'), env=self.environment, shell=True)
            utils.execute('rpm -qa > "/tmp/%s.txt"' % _('Package list'), \
env=self.environment, shell=True)
        # Disk info?
        if self.diskinfo == '1':
            self.ifCancel()
            utils.execute('fdisk -l > "/tmp/%s.txt"' % _('Disk Information'), \
env=self.environment, shell=True)

        # IF and only IF tar / tar.gz, add .tar to since
        # it's a file, not a folder the full destination
        if self.engine.startswith('tar'):
            self.backupdest += '.tar'

        # Commands, etc based on engine
        if self.engine.startswith('tar'):
            command = 'tar rf \'%s\'' % (self.backupdest)
            os.chdir('/tmp')
            if self.pkglist == '1':
                utils.execute('%s "/tmp/%s.txt" "/tmp/%s.txt"' % (command, \
_('Package list'), _('Package list, names only')), env=self.environment, \
shell=True, timeout=-1)
                utils.execute('rm "/tmp/%s.txt" "/tmp/%s.txt" -f' % \
(_('Package list'), _('Package list, names only')), env=self.environment, \
shell=True, timeout=-1)
            if self.diskinfo == '1':
                utils.execute('%s "/tmp/%s.txt"' % (command, _('Disk Information')),\
                    env=self.environment, shell=True, timeout=-1)
                utils.execute('rm "/tmp/%s.txt" -f' % _('Disk Information'),\
                    env=self.environment, shell=True, timeout=-1)
            if not self.recursive == '1':
                command += ' --no-recursion'
            if not self.dohidden == '1':
                command += ' --exclude=.*'

        elif self.engine == 'rsync':
            # Create backup dir first
            os.mkdir(self.backupdest)
            command = 'rsync -g -o -p'
            if self.recursive == '1':
                command += ' -r'
            if not self.dohidden == '1':
                command += ' --exclude=.*'
            if self.pkglist == '1':
                self.ifCancel()
                retval, a, b = utils.execute('%s "/tmp/%s.txt" "/tmp/%s.txt" "%s"' \
% (command, _('Package list'),\
                    _('Package list, names only'), self.backupdest), \
env=self.environment, shell=True, timeout=-1)
                utils.execute('rm "/tmp/%s.txt" "/tmp/%s.txt" -f' % (_('Package list'),\
                    _('Package list, names only')), env=self.environment, shell=True, timeout=-1)
            if self.diskinfo == '1':
                self.ifCancel()
                utils.execute('%s "/tmp/%s.txt" "%s"' % (command, \
_('Disk Information'), self.backupdest), env=self.environment, shell=True, timeout=-1)
                utils.execute('rm "/tmp/%s.txt" -f' % _('Disk Information'),\
                    env=self.environment, shell=True, timeout=-1)

        # Get the list of paths...
        paths = []
        for i in self.set.options('Paths'):
            paths.append('\'%s\'' % self.set.get('Paths', i))
            self.total = self.total + 1

        # Now that the paths & commands are set up...
        for i in paths:
            self.ifCancel()
            self.current = self.current + 1
            self.logger.logmsg('DEBUG', _('Backing up path \'%s\'') % i.split('\'')[1])
            if self.engine == 'rsync':
                i = '%s "%s"' % (i, self.backupdest)
            retval, stdout, stderr = utils.execute('%s %s' % (command, i), \
env=self.environment, shell=True)
            if retval != os.EX_OK and retval != 2:
                self.logger.logmsg('ERROR', _('Error backing up path \'%s\'') % i)
                self.logger.logmsg('ERROR', _('Output:\n%s') % ('Return value: %s (Maybe the process was killed?)\\nstdout: %s\nstderr: %s' %\
                    (retval, stdout.readlines(), stderr.readlines())))

        # gzip at end if needed
        if self.engine == 'tar.gz':
            self.ifCancel()
            self.logger.logmsg('INFO', _('Compressing the backup archive...'))
            retval, stdout, stderr = utils.execute('gzip \'%s\''\
                                                    % self.backupdest, \
                                                    env=self.environment, \
                                                    shell=True)
            if retval != os.EX_OK:
                self.logger.logmsg('WARNING', _('An error occured while \
compressing the backup. Your backup is complete, but may not be (completely) \
compressed.'))
                self.logger.logmsg('ERROR', _('Output:\n%s') % ('Return value: \
%s\nstdout: %s\nstderr: %s' %\
                    (retval, stdout.readlines(), stderr.readlines())))

        # Backups are private!
        if self.engine != 'tar.gz':
            os.chmod(self.backupdest, 0711)
        else:
            os.chmod('%s.gz' % self.backupdest, 0711)

        # After command
        if self.command_after != None and self.command_after != "":
            self.ifCancel()
            self.logger.logmsg('INFO', _('Executing \'After\' command'))
            retval, stdout, stderr = utils.execute(\
                                                   self.command_after, \
                                                   env=self.environment, \
                                                   shell=True\
                                                   )
            if retval != os.EX_OK:
                self.logger.logmsg('WARNING', _('Command failed; Please check \
your syntax to make sure it is correct.'))

        # All done!
        self.logger.logmsg('INFO', _('Finished automatic backup of set \'%s\'')\
                            % self.setname)

