# -*- coding: utf-8 -*-
#    Copyright (C) 2005, 2006, 2007 Stewart Adam
#    This file is part of fwbackups.

#    fwbackups is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    (at your option) any later version.

#    fwbackups is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.

#    You should have received a copy of the GNU General Public License
#    along with fwbackups; if not, write to the Free Software
#    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

# A lot of the cursor stuff based on Yumex - Thanks!
# Original author:    Tim Lauridsen <tla@rasmil.dk>
"""
Extra functions for fwbackups
"""

import os, os.path
import sys
import subprocess, subprocess_killable
from i18n import _
import traceback
import exceptions
import logging
from threading import Thread,Event

# This is the only var NOT in const to avoid import dep issues
LOGGERNAME='fwbackups-main'

def ConvertPath(path):
    """ Converts Linux-naitive paths into portable ones. 
        path: Path to convert
    """
    return os.sep.join(path.split('/'))

def Portability(user):
    """ Returns the full paths to all locations used in fwbackups,
        plus a bit of portability stuff based on the user given.
        Returns:
            pathsep: same as os.sep
            linesep: same as os.linesep
            userhome: The full path to the user's home
            loc: The fwbackups configuration directory
            setloc: The location to store set configuration files
            onetimeloc: The location to store the one-time backup configuration file
            prefsloc: The location of the preferences file
            logloc: The location to store the log file
            To use: (pathsep, linesep, userhome, loc, setloc, onetimeloc, prefsloc, logloc) = Portability()
    """
    pathsep = os.sep
    linesep = os.linesep
    userhome = os.path.expanduser('~%s' % user)
    loc = ConvertPath('%s/.fwbackups' % userhome)
    setloc = ConvertPath('%s/.fwbackups/Sets' % userhome)
    onetimeloc = ConvertPath('%s/fwbackups-OneTime.conf' % loc)
    prefsloc = ConvertPath('%s/fwbackups-preferences.conf' % loc)
    logloc = ConvertPath('%s/fwbackups-userlog.txt' % loc)
    return pathsep, linesep, userhome, loc, setloc, onetimeloc, prefsloc, logloc

def execute(command, env=None, shell=False, timeout=-1):
    """ Execute a command, wait for it to finish """
    sub = subprocess_killable.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell, env=env)
    return sub.wait(timeout), sub.stdout, sub.stderr

def executeSub(command, env=None, shell=False):
    """ Execute a command async """
    sub = subprocess_killable.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=shell, env=env)
    return sub

def CheckPerms(path, recursive=False):
    """ Checks for read and write permissions.
        Takes the path to check as an argument, returns
        True if permissions check out, False if not.
    """
    read = CheckPermsRead(path, recursive)
    write = CheckPermsWrite(path, recursive)
    if read == True and write == True:
        return True
    return False
# ideally, use below but wtvr.
"""def CheckPerms(path, recursive=False):
    \""" Checks for read and write permissions.
        Takes the path to check as an argument, returns
        True if permissions check out, False if not.
    \"""
    if path.endswith('/'):
        path2 = '/'.join(path.split('/')[:-1])
    else:
        path2 = path
    if not os.access(path, os.R_OK|os.W_OK):
        return False
    if recursive:
        for i in os.listdir(path):
            fullpath = '%s%s%s' % (path2, os.sep, i)
            if os.path.isdir(fullpath) and fullpath != '/proc' and fullpath != '/sys'\
                and os.access(fullpath, os.R_OK) and not os.path.islink(fullpath):
                CheckPerms(fullpath, True)
            elif os.path.isfile(fullpath) and not os.access(fullpath, os.R_OK|os.W_OK):
                return False
    else:
        if not os.access(path, os.R_OK|os.W_OK):
            return False
    return True"""

def CheckPermsRead(path, recursive=False):
    """ Checks for read permissions.
        Takes the path to check as an argument, returns
        True if permissions check out, False if not.
    """
    if path.endswith('/'):
        path2 = '/'.join(path.split('/')[:-1])
    else:
        path2 = path
    if not os.access(path, os.R_OK):
        return False
    if recursive:
        for i in os.listdir(path):
            fullpath = '%s%s%s' % (path2, os.sep, i)
            if os.path.isdir(fullpath) and fullpath != '/proc' and fullpath != '/sys'\
                and os.access(fullpath, os.R_OK) and not os.path.islink(fullpath):
                CheckPerms(fullpath, True)
            elif os.path.isfile(fullpath) and not os.access(fullpath, os.R_OK):
                return False
    else:
        if not os.access(path, os.R_OK):
            return False
    return True

def CheckPermsWrite(path, recursive=False):
    """ Checks for write permissions.
        Takes the path to check as an argument, returns
        True if permissions check out, False if not.
    """
    if path.endswith('/'):
        path2 = '/'.join(path.split('/')[:-1])
    else:
        path2 = path
    if not os.access(path, os.W_OK):
        return False
    if recursive:
        for i in os.listdir(path):
            fullpath = '%s%s%s' % (path2, os.sep, i)
            if os.path.isdir(fullpath) and fullpath != '/proc' and fullpath != '/sys'\
                and os.access(fullpath, os.W_OK) and not os.path.islink(fullpath):
                CheckPerms(fullpath, True)
            elif os.path.isfile(fullpath) and not os.access(fullpath, os.W_OK):
                return False
    else:
        if not os.access(path, os.W_OK):
            return False
    return True

class FuncAsThread(Thread):
    """ Run a function as a new thread. """
    def __init__(self, functorun):
        Thread.__init__(self)
        self.functorun = functorun
        
    def run(self):
        try:
            retval = self.functorun()
            return retval
        except exceptions.SystemExit:
            sys.exit(0)
        except: # catch other exception and write it to the logger.
            logger = logging.getLogger(LOGGERNAME)
            etype = sys.exc_info()[0]
            evalue = sys.exc_info()[1]
            etb = traceback.extract_tb(sys.exc_info()[2])
            logger.logmsg('ERROR', _('Error while running the operation\'s subprocess.'))
            logger.logmsg('ERROR', _('Error Type: ' + str(etype)))
            logger.logmsg('ERROR', _('Error Value: ' + str(evalue)))
            logger.logmsg('ERROR', _('Traceback: \n'))
            for tub in etb:
                logger.error(str(tub))
                f,l,m,c = tub # file,lineno, function, codeline
                logger.error(_('File : %s , line %s, in %s') % (f,str(l),m))
                logger.error('    %s ' % c)
            self.exception = etype(evalue)
            sys.exit(1)
